"""
This module contains functionality for managing the ruleset on the
PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> r = pl.Ruleset()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import resource
from thisversion import PLDB
from shared import shared

import netobject
import portobject
import vlanobject
import serviceobject
import protocolobject
import aspathobject
import timeobject
import propertyobject
import rewriteobject
import flagobject
import systemobject
import mplsobject
import dscpobject
import channelobject

import shapingobject
import statisticsobject

import plobject
import plrule

import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

class _Cache:
	def __init__(self, root):
		"""
		Object containing the local cache for PLobjects.
		
		Each object type (e.g 'NetObjects') will have it's own instance of this class.

		The <paths> index is a dictionary with the fullpath as key, and the value is
		another dict. That dict will have each systemid as a key and the value is
		a <node> tuple. There is also a duplicate None key for local objects, that
		points to the same <node> as it's systemid.

		Each <node>, which exists per path/systemid, is a tuple where the first item
		is the Object, and the secont item a list of all children in that systemid.
		Each child will be the <node> for that child. (The list is empty if there
		are no children.)

		The <root> parameter is e.g '/NetObjects' and there will be dictionary at that
		path, though the <node> for each systemid there will have None as object.

		The <ids> index is a dictionary with one key for each systemid, and an extra
		key None for local objects, and the value of each of them is another dict
		with the id as key and the <node> as value.

		Adding an object with the add() method require that it's parent is present
		in the cache. (So we cant iterate a list and do add on the items, unless we
		know the list is sorted with parents before it's children.)
		"""
		self.root = root
		self.ids = {None: {}}
		self.paths = {root: {}}
		##print "Init cache for %s" % root

	def create(self, unsorted):
		""" Create the cache content by adding the objects in the <unsorted> list."""
		# This is not an optimised way to do it. But adding shortest paths first ensures that parents are present.
		##print "Sorting list of objects..."
		lenlist = [(len(obj.fullpath), obj) for obj in unsorted]
		lenlist.sort()
		##print "Adding objects to cache..."
		for l, obj in lenlist:
			self.add(obj)

	def _sids(self, obj):
		""" Return a list of [obj.systemid] if is_global is true,
		or [obj.systemid, None] if is_global is false.
		"""
		if obj.is_global:
			return [obj.systemid]

		else:
			return [obj.systemid, None]
		
	def add(self, obj):
		""" Add an object to the cache. Parent must exist."""
		parent = self.paths[obj.path]
		for sid in self._sids(obj):
			##print "Adding %s for sid %s" % (obj.fullpath, sid)
			# Create a new node for this path/systemid
			node = (obj, [])

			# Add to paths and ids index
			self.paths.setdefault(obj.fullpath, {})[sid] = node
			self.ids.setdefault(sid, {})[obj.id] = node

			# Add to parent nodes list of children
			if not obj.parent and sid not in parent:
				parent[sid] = (None, [])
				
			parent[sid][1].append(node)

	def remove(self, obj):
		""" Remove an object from the cache."""
		for sid in self._sids(obj):
			node = self.paths[obj.fullpath][sid]
		
			# Remove any children on this systemid
			self._recursive_remove(node[1], sid)
			
			# Remove node from indexes
			del self.ids[sid][obj.id]
			del self.paths[obj.fullpath][sid]
			
			# Remove node from parent
			self.paths[obj.path][sid][1].remove(node)

		# If there are no other systemid:s at this path, remove it completely
		if not self.paths[obj.fullpath]:
			del self.paths[obj.fullpath]

	def _recursive_remove(self, children, sid):
		""" Remove child objects from the cache recursively."""
		for node in children:
			self._recursive_remove(node[1], sid)
			del self.ids[sid][node[0].id]
			del self.paths[node[0].fullpath][sid]
			if not self.paths[node[0].fullpath]:
				del self.paths[node[0].fullpath]
			
	def move(self, obj, new_parent):
		""" Move an object in the cache to a new path.
		It also updates the objects (full)path properties.
		"""
		assert new_parent in self.paths, "New parent does not exist"
		assert obj.path != new_parent, "Source and destination path is the same."
		assert not new_parent.startswith(obj.fullpath + "/"), "You can't move an object into it's own child path"
		assert obj.fullpath in self.paths
		assert obj.systemid in self.ids and obj.id in self.ids[obj.systemid]

		new_fullpath = new_parent + "/" + obj.name
		old_offset = len(obj.path) + 1
		
		# Move the node
		for sid in self._sids(obj):
			node = self.paths[obj.fullpath][sid]

			# Create a new path pointer to the object, and remove the old one.
			self.paths.setdefault(new_fullpath, {})[sid] = node #self.paths[obj.fullpath][sid]
			del self.paths[obj.fullpath][sid]

			# Remove from old parents list of children, and append to new parents list
			self.paths[obj.path][sid][1].remove(node)
			self.paths[new_parent][sid][1].append(node)
			
			# Loop all the children and update the paths index for them
			self._recursive_move(node[1], sid, new_parent + "/", old_offset)

		# If there are no other systemid left at this path, remove it.
		if not self.paths[obj.fullpath]:
			del self.paths[obj.fullpath]

		# Update the (full)path properties of the object, and children
		obj._set_path(new_fullpath)
		self._recursive_setpath(self.paths[new_fullpath][obj.systemid][1], new_parent + "/", old_offset)
		
	def _recursive_move(self, nodes, sid, new_parent_slash, old_offset):
		""" Recursively move children to a new path."""
		for node in nodes:
			obj, children = node
			self._recursive_move(children, sid, new_parent_slash, old_offset)
			self.paths.setdefault(new_parent_slash + obj.fullpath[old_offset:], {})[sid] = node
			del self.paths[obj.fullpath][sid]
			if not self.paths[obj.fullpath]:
				del self.paths[obj.fullpath]
				
	def _recursive_setpath(self, nodes, new_parent_slash, old_offset):
		""" Recursively set the (full)path properties of objects, after a move."""
		for obj, children in nodes:
			obj._set_path(new_parent_slash + obj.fullpath[old_offset:])
			self._recursive_setpath(children, new_parent_slash, old_offset)
			
	def find_path(self, path, systemid=None):
		""" Find an object by full path + name, in specified systemid or local objects.
		Returns None if not found.
		"""
		return self.paths.get(path, {}).get(systemid, (None,))[0]

	def find_id(self, id, systemid=None):
		""" Find an object by ID number, in specified systemid or local objects.
		Returns None if not found.
		"""
		return self.ids.get(systemid, {}).get(id, (None,))[0]

	def list(self, path=None, recursive=True, systemid=None):
		""" List objects from given path, or from root.
		List only specified systemid, or any system.
		List specified objects and all children recursively,
		or just first level children of specified object.
		Returns an empty list if nothing was found.
		"""
		if path is None:
			path = self.root
			
		if path not in self.paths:
			# This path does not exist at all - abort
			return []

		nodes = self.paths[path]
		if systemid is not None and systemid not in nodes:
			# This path does not exist for this systemid - abort
			return []

		ret = []
		if not recursive:
			# We only want the first level children to the specified path
			if systemid is None:
				# Loop through each systemid (except the duplicate None)
				for sid, node in nodes.items():
					if sid is not None:
						for object, children in node[1]:
							ret.append(object)

			else:
				# Only fetch the specified systemid
				for object, children in nodes[systemid][1]:
					ret.append(object)
					
			return ret

		# Recursive, return the specified object, and children at all levels
		if systemid is None:
			# Loop through each systemid (except the duplicate None)
			for sid, (object, children) in nodes.items():
				if sid is not None:
					if object is not None:
						ret.append(object)
						
					self._recursive_list(ret, children)

		else:
			# Only fetch the specified systemid
			object, children = nodes[systemid]
			if object is not None:
				ret.append(object)
						
			self._recursive_list(ret, children)

		return ret

	def _recursive_list(self, ret, nodes):
		""" Append all children objects to <ret> list recursively."""
		for object, children in nodes:
			ret.append(object)
			self._recursive_list(ret, children)
		
class Ruleset(resource.Resource):

	"""
	The Ruleset resource is used for reading and manipulating
	objects and rules in the PacketLogic ruleset.

	See also 
	L{NetObjects <netobject.NetObject>},
	L{PortObjects <portobject.PortObject>},
	L{ProtocolObjects <protocolobject.ProtocolObject>},
	L{ServiceObjects <serviceobject.ServiceObject>},
	L{TimeObjects <timeobject.TimeObject>},
	L{VlanObjects <vlanobject.VLANObject>},
	L{PropertyObjects <propertyobject.PropertyObject>},
	L{RewriteObjects <rewriteobject.RewriteObject>},
	L{FlagObjects <flagobject.FlagObject>} and 
	L{SystemObjects <systemobject.SystemObject>},
	L{MPLSObjects <mplsobject.MPLSObject>},
	L{DSCPObjects <mplsobject.DSCPObject>},
	L{ChannelObjects <mplsobject.ChannelObject>},
	L{ShapingObjects <shapingobject.ShapingObject>},
	L{StatisticsObjects <statisticsobject.StatisticsObject>},
	as well as
	L{FwRules <plrule.FwRule>} and
	L{ShapingRules <plrule.ShapingRule>}.
	L{StatisticsRules <plrule.StatisticsRule>}.

	@group Objects: object_add, object_find, object_find_id, object_get, object_list, object_remove, object_unlink

	@group Shaping objects: shapingobject_list, shapingobject_remove, shapingobject_add, shapingobject_find, shapingobject_find_id

	@group Statistics objects: statisticsobject_list, statisticsobject_remove, statisticsobject_add, statisticsobject_find, statisticsobject_find_id
	
	@group Firewall rules: fwrule_add, fwrule_find, fwrule_find_id, fwrule_list, fwrule_remove
	@group Shaping rules: shapingrule_add, shapingrule_find, shapingrule_find_id, shapingrule_list, shapingrule_remove
	@group Statistics rules: statisticsrule_add, statisticsrule_find, statisticsrule_find_id, statisticsrule_list, statisticsrule_remove

	@group Services: services_cmd, services_list, services_prop_list
	@group Deprecated: add, remove, unlink

	@group Condition types: CONDITION_NETOBJECT_CLIENT, CONDITION_NETOBJECT_SERVER, CONDITION_NETOBJECT_HOST, CONDITION_PORTOBJECT_CLIENT, CONDITION_PORTOBJECT_SERVER, CONDITION_SERVICEOBJECT, CONDITION_VLANOBJECT, CONDITION_VLANOBJECT_IN, CONDITION_VLANOBJECT_OUT, CONDITION_PROTOCOLOBJECT, CONDITION_TIMEOBJECT, CONDITION_ASPATHOBJECT, CONDITION_PROPERTYOBJECT, CONDITION_SYSTEMOBJECT, CONDITION_FLAGOBJECT, CONDITION_MPLSOBJECT, CONDITION_MPLSOBJECT_IN, CONDITION_MPLSOBJECT_OUT, CONDITION_NETOBJECT_LOCAL, CONDITION_DSCPOBJECT, CONDITION_DSCPOBJECT_IN, CONDITION_DSCPOBJECT_OUT, CONDITION_CHANNELOBJECT, CONDITION_CHANNELOBJECT_IN, CONDITION_CHANNELOBJECT_OUT
	@group Condition operators: CONDITION_OP_EQ, CONDITION_OP_NE
	@group Firewall actions: FWRULE_ACTION_ACCEPT, FWRULE_ACTION_REJECT, FWRULE_ACTION_DROP, FWRULE_ACTION_REWRITE, FWRULE_ACTION_DIVERT, FWRULE_ACTION_INJECT
	@group Split by type: SPLIT_NONE, SPLIT_LOCALHOST, SPLIT_HOST_NETOBJECT, SPLIT_SERVER_NETOBJECT, SPLIT_CLIENT_NETOBJECT, SPLIT_CONNECTION

	@cvar CONDITION_NETOBJECT_CLIENT: Client NetObjects
	@cvar CONDITION_NETOBJECT_SERVER: Server NetObjects
	@cvar CONDITION_NETOBJECT_HOST: Host NetObjects
	@cvar CONDITION_NETOBJECT_LOCAL: Local NetObjects
	@cvar CONDITION_PORTOBJECT_CLIENT: Client PortObjects
	@cvar CONDITION_PORTOBJECT_SERVER: Server PortObjects
	@cvar CONDITION_SERVICEOBJECT: ServiceObjects
	@cvar CONDITION_VLANOBJECT: VLANObjects
	@cvar CONDITION_VLANOBJECT_IN: Inbound VLANObjects
	@cvar CONDITION_VLANOBJECT_OUT: Outbound VLANObjects
	@cvar CONDITION_PROTOCOLOBJECT: ProtocolObjects
	@cvar CONDITION_TIMEOBJECT: TimeObjects
	@cvar CONDITION_ASPATHOBJECT: ASPathObjects
	@cvar CONDITION_PROPERTYOBJECT: PropertyObjects
	@cvar CONDITION_SYSTEMOBJECT: SystemObjects
	@cvar CONDITION_FLAGOBJECT: FlagObjects
	@cvar CONDITION_MPLSOBJECT: MPLSObjects
	@cvar CONDITION_MPLSOBJECT_IN: Inbound MPLSObjects
	@cvar CONDITION_MPLSOBJECT_OUT: Outbound MPLSObjects
	@cvar CONDITION_DSCPOBJECT: DSCPObjects
	@cvar CONDITION_DSCPOBJECT_IN: Inbound DSCPObjects
	@cvar CONDITION_DSCPOBJECT_OUT: Outbound DSCPObjects
	@cvar CONDITION_CHANNELOBJECT: CHANNELObjects
	@cvar CONDITION_CHANNELOBJECT_IN: Inbound CHANNELObjects
	@cvar CONDITION_CHANNELOBJECT_OUT: Outbound CHANNELObjects

	@cvar CONDITION_OP_EQ: Condition equals
	@cvar CONDITION_OP_NE: Condition does not equal

	@cvar FWRULE_ACTION_ACCEPT: Accept connection
	@cvar FWRULE_ACTION_REJECT: Reject connection
	@cvar FWRULE_ACTION_DROP: Drop connection
	@cvar FWRULE_ACTION_REWRITE: Rewrite connection
	@cvar FWRULE_ACTION_DIVERT: Divert connection
	@cvar FWRULE_ACTION_INJECT: Inject data into connection

	@cvar SPLIT_NONE: Split by none
	@cvar SPLIT_LOCALHOST: Split by localhost
	@cvar SPLIT_HOST_NETOBJECT: Split by host netobject
	@cvar SPLIT_LOCAL_NETOBJECT: Split by local netobject
	@cvar SPLIT_SERVER_NETOBJECT: Split by server netobject
	@cvar SPLIT_CLIENT_NETOBJECT: Split by client netobject
	@cvar SPLIT_CONNECTION: Split by connection
	@cvar SPLIT_SUBSCRIBER: Split by subscriber
	"""

	SPLIT_NONE = PLDB.SHAPINGOBJECT_SPLIT_NONE
	SPLIT_LOCALHOST = PLDB.SHAPINGOBJECT_SPLIT_LOCALHOST
	SPLIT_HOST_NETOBJECT = PLDB.SHAPINGOBJECT_SPLIT_HOST_NETOBJECT
	SPLIT_SERVER_NETOBJECT = PLDB.SHAPINGOBJECT_SPLIT_SERVER_NETOBJECT
	SPLIT_CLIENT_NETOBJECT = PLDB.SHAPINGOBJECT_SPLIT_CLIENT_NETOBJECT
	SPLIT_CONNECTION = PLDB.SHAPINGOBJECT_SPLIT_CONNECTION
	SPLIT_LOCAL_NETOBJECT = shapingobject.NONEXISTANT_SPLIT_TYPE
	SPLIT_SUBSCRIBER = shapingobject.NONEXISTANT_SPLIT_TYPE

	FWRULE_ACTION_ACCEPT = PLDB.FWRULE_ACTION_ACCEPT
	FWRULE_ACTION_REJECT = PLDB.FWRULE_ACTION_REJECT
	FWRULE_ACTION_DROP = PLDB.FWRULE_ACTION_DROP
	FWRULE_ACTION_REWRITE = PLDB.FWRULE_ACTION_REWRITE
	FWRULE_ACTION_DIVERT = PLDB.FWRULE_ACTION_DIVERT
	FWRULE_ACTION_INJECT = plrule.NONEXISTANT_FWRULE_ACTION

	CONDITION_OP_EQ = PLDB.RULESET_CONDITION_OP_EQ
	CONDITION_OP_NE = PLDB.RULESET_CONDITION_OP_NE

	CONDITION_NETOBJECT_CLIENT = PLDB.RULESET_CONDITION_TYPE_NETOBJECT_CLIENT
	CONDITION_NETOBJECT_SERVER = PLDB.RULESET_CONDITION_TYPE_NETOBJECT_SERVER
	CONDITION_NETOBJECT_HOST = PLDB.RULESET_CONDITION_TYPE_NETOBJECT_HOST
	CONDITION_NETOBJECT_LOCAL = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_PORTOBJECT_CLIENT = PLDB.RULESET_CONDITION_TYPE_PORTOBJECT_CLIENT
	CONDITION_PORTOBJECT_SERVER = PLDB.RULESET_CONDITION_TYPE_PORTOBJECT_SERVER
	CONDITION_SERVICEOBJECT = PLDB.RULESET_CONDITION_TYPE_SERVICEOBJECT
	CONDITION_VLANOBJECT = PLDB.RULESET_CONDITION_TYPE_VLANOBJECT
	CONDITION_VLANOBJECT_IN = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_VLANOBJECT_OUT = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_PROTOCOLOBJECT = PLDB.RULESET_CONDITION_TYPE_PROTOCOLOBJECT
	CONDITION_TIMEOBJECT = PLDB.RULESET_CONDITION_TYPE_TIMEOBJECT
	CONDITION_ASPATHOBJECT = PLDB.RULESET_CONDITION_TYPE_ASPATHOBJECT
	CONDITION_PROPERTYOBJECT = PLDB.RULESET_CONDITION_TYPE_PROPERTYOBJECT
	CONDITION_SYSTEMOBJECT = PLDB.RULESET_CONDITION_TYPE_SYSTEMOBJECT
	CONDITION_FLAGOBJECT = PLDB.RULESET_CONDITION_TYPE_FLAGOBJECT
	CONDITION_MPLSOBJECT = PLDB.RULESET_CONDITION_TYPE_MPLSOBJECT
	CONDITION_MPLSOBJECT_IN = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_MPLSOBJECT_OUT = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_DSCPOBJECT = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_DSCPOBJECT_IN = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_DSCPOBJECT_OUT = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_CHANNELOBJECT = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_CHANNELOBJECT_IN = plrule.NONEXISTANT_CONDITION_TYPE
	CONDITION_CHANNELOBJECT_OUT = plrule.NONEXISTANT_CONDITION_TYPE

	def __init__(self, host, user, password, port=42000, timeout=None):
		"""
		@see: L{Resource <resource.Resource.__init__>}.
		
		@type host: C{str}
		@type user: C{str}
		@type password: C{str}
		"""
		resource.Resource.__init__(self, host, user, password, port=port, timeout=timeout)
		self.objects = {"NetObjects":PLDB.OBJECT_TYPE_NETOBJECT,
				"PortObjects":PLDB.OBJECT_TYPE_PORTOBJECT,
				"VLANObjects":PLDB.OBJECT_TYPE_VLANOBJECT,
				"ServiceObjects":PLDB.OBJECT_TYPE_SERVICEOBJECT,
				"ASPathObjects":PLDB.OBJECT_TYPE_ASPATHOBJECT,
				"TimeObjects":PLDB.OBJECT_TYPE_TIMEOBJECT,
				"PropertyObjects":PLDB.OBJECT_TYPE_PROPERTYOBJECT,
				"RewriteObjects":PLDB.OBJECT_TYPE_REWRITEOBJECT,
				"ProtocolObjects":PLDB.OBJECT_TYPE_PROTOCOLOBJECT,
				"FlagObjects":PLDB.OBJECT_TYPE_FLAGOBJECT,
				"SystemObjects":PLDB.OBJECT_TYPE_SYSTEMOBJECT,
				"MPLSObjects":PLDB.OBJECT_TYPE_MPLSOBJECT}
		self.type_to_obj = {PLDB.OBJECT_TYPE_NETOBJECT:netobject.NetObject,
				    PLDB.OBJECT_TYPE_PORTOBJECT:portobject.PortObject,
				    PLDB.OBJECT_TYPE_VLANOBJECT:vlanobject.VLANObject,
				    PLDB.OBJECT_TYPE_SERVICEOBJECT:serviceobject.ServiceObject,
				    PLDB.OBJECT_TYPE_ASPATHOBJECT:aspathobject.ASPathObject,
				    PLDB.OBJECT_TYPE_TIMEOBJECT:timeobject.TimeObject,
				    PLDB.OBJECT_TYPE_PROPERTYOBJECT:propertyobject.PropertyObject,
				    PLDB.OBJECT_TYPE_REWRITEOBJECT:rewriteobject.RewriteObject,
				    PLDB.OBJECT_TYPE_PROTOCOLOBJECT:protocolobject.ProtocolObject,
				    PLDB.OBJECT_TYPE_FLAGOBJECT:flagobject.FlagObject,
				    PLDB.OBJECT_TYPE_SYSTEMOBJECT:systemobject.SystemObject,
				    PLDB.OBJECT_TYPE_MPLSOBJECT:mplsobject.MPLSObject}
				    
		self._objectcaches = {} ###

		self._fwrulecache = None
		self._fwrulecache_id = None
		self._fwrulecache_name = None

		self._shapingobjectcache = None
		self._shapingobjectcache_id = None
		self._shapingobjectcache_name = None

		self._shapingrulecache = None
		self._shapingrulecache_id = None
		self._shapingrulecache_name = None

	def _resource(self):
		return PLDB.RESOURCE_RULESET

	def _split_pathstr(self, pathstr, allowEmpty=True):

		obj = None
		path = None

		if len(pathstr) > 1:
			s = pathstr.split("/")
			if len(s) > 1:
				obj = s[1]

			if len(s) > 2:
				path = s[2:]

		if not allowEmpty:
			if not path or "" in path:
				raise ValueError("Invalid path '%s'" % pathstr)

	#	if not self.objs.has_key(obj):
	#		raise ValueError("Path is invalid. Object %s doesn't exist" % object)

		return (obj, path)


	def _split_pathstr_check(self, pathstr, allowEmpty=True):
		obj, path = self._split_pathstr(pathstr, allowEmpty=allowEmpty)
		if not self.objects.has_key(obj):
			raise ValueError("Object type '%s' does not exist" % obj)
		return (obj, path)

	def _has_feature(self, feature):
		""" Return True or False depending on if the feature is enabled."""
		if not hasattr(shared, 'features'):
			shared.features = self.features_get()

		return getattr(shared.features, feature)
		
	def features_get(self):
		"""Returns a simple object with boolean attributes showing enabled features.
		The members of the object, that can be True or False, are:
		bgp, fw, shaping, statistics, connlog (and its alias connsearch),
		enterprise, monitor, divert and vbs

		@returns: Object with a boolean attribute for each feature.
		@rtype: object
		
		"""
		class Features:
			def __repr__(self):
				return "<Features [%s]>" % (", ".join([n for n in dir(self) if not n.startswith('_') and getattr(self, n)]))
			
		bits = plutil._do(self, self.session.features_get(), plutil._AddCB(self, self.session.features_get_handle))
		features = Features()
		features.bgp = bool(bits & PLDB.FEATURES_BGP)
		features.fw = bool(bits & PLDB.FEATURES_FW)
		features.shaping = bool(bits & PLDB.FEATURES_SHAPING)
		features.statistics = bool(bits & PLDB.FEATURES_STATISTICS)
		features.connlog = bool(bits & PLDB.FEATURES_CONNSEARCH)
		features.connsearch = bool(bits & PLDB.FEATURES_CONNSEARCH)
		features.enterprise = bool(bits & PLDB.FEATURES_ENTERPRISE)
		features.monitor = bool(bits & PLDB.FEATURES_MONITOR)
		features.divert = bool(bits & PLDB.FEATURES_DIVERT)
		features.vbs = bool(bits & PLDB.FEATURES_VBS)
		
		return features
		
	def refresh(self, force=False, timeout=1.0):
		""" Check for any pending "New data was committed" signals, and
		invalidate cached objects and rules if it occurs, or force
		them to be invalidated with force=True.
		Returns True if such an exception occured (and was trapped),
		otherwise false. Any other exception than the PLDBError "New
		data comitted" or timeout will not be trapped.

			>>> rs.refresh(); rs.object_add('/NetObjects/TestObject'); rs.commit()
			True
			<PacketLogic NetObject 'TestObject' '/NetObjects'>

		@type force: C{bool}
		@param force: If set to True the cache in the API of rules
		     and objects will be marked dirty, forcing them to be
		     reread from the PacketLogic system upon the next query,
		     even if no "New data was comitted" signal occured.
		
		@type timeout: C{float}
		@param timeout: Seconds to wait for any data to arrive from
		     the PacketLogic. Since we typically check for "New data
		     was commited" that already are pending, this can be
		     short, and default is one second.
		     Zero is not allowed, and will silently be changed to 0.1

		@rtype: C{bool}
		@returns: True if a "New data was committed" signal was seen,
		     otherwise False.
		"""
		
		if timeout < 0.1:
			# Setting to 0 will change the (blocking) characteristict of the
			# socket, and we prevent that with forcing at least 0.1 seconds.
			timeout = 0.1
			
		ret = False
		to = self.timeout
		self.timeout = timeout

		while True:
			try:
				self._process()
				# We actually will never exit without an exception, since we have no command/handler.
				break
			
			except IOError, msg:
				if not str(msg) == "Operation timed out":
					self.timeout = to
					raise
				break
				
			except resource.PLDBError, msg:
				if not 'New data was committed' in str(msg):
					self.timeout = to
					raise

				ret = True

		self.timeout = to
		
		if force or ret:
			self._invalidate_cache()
			
		return ret
		
	"""
	Object cache code
	"""

	def _invalidate_cache(self, cache_type=None):
		self._fwrule_invalidate_cache()
		self._shapingrule_invalidate_cache()
		self._shapingobject_invalidate_cache()

		if cache_type is None:
			self._objectcaches.clear() ###
		elif self._objectcaches.has_key(cache_type):
			del self._objectcaches[cache_type] ###

	def _update_cache(self, cache_type):
		### @raises RuntimeError("Unknown object item type"): for invalid type of object. Should never happen unless e.g two api-versions are mixed up.
		### FIXME supply textual root to use for cache instansiation, and optimise (away) makepath/iddict, better create() perhaps
		if self._objectcaches.has_key(cache_type):
			return

		l = plobject._list(self, cache_type)

		iddict = dict([((o.systemid, o.id), o) for o in l])
		for o in l:
			o._make_path(iddict)
			
		self._objectcaches[cache_type] = _Cache(self.type_to_obj[cache_type]().base_path) ###
		self._objectcaches[cache_type].create(l) ###

	def object_list(self, path="/", recursive=True, systemid=None):
		"""
		Lists all the Objects that match the path.

			>>> r.object_list('/NetObjects')
			[<PacketLogic NetObject 'Netaccess Kunder' '/NetObjects'>, <PacketLogic NetObject 'Netaccess allowed servers' '/NetObjects'>,
			<PacketLogic NetObject 'nxs' '/NetObjects'>, <PacketLogic NetObject 'DummyPlaceholder' '/NetObjects'>,
			<PacketLogic NetObject 'nxs_Bandbredd 2000' '/NetObjects/nxs'>, <PacketLogic NetObject 'TestObject' '/NetObjects'>,
			<PacketLogic NetObject 'TestNetobject' '/NetObjects'>, <PacketLogic NetObject 'TestNetobject-Deuxe' '/NetObjects'>,
			<PacketLogic NetObject 'NetintactTest' '/NetObjects'>, <PacketLogic NetObject 'Kruken' '/NetObjects/nxs'>,
			<PacketLogic NetObject 'nxs_1M [3]' '/NetObjects/nxs'>, <PacketLogic NetObject 'nxs_4Mbit [2]' '/NetObjects/nxs'>,
			<PacketLogic NetObject 'nxs_256kbit [1]' '/NetObjects/nxs'>]

		@param path: The path of Objects you want to list. I.e. you
			want to list all the L{NetObjects <netobject.NetObject>} set the path to
			I{/NetObjects}. This will recursively list all L{NetObjects <netobject.NetObject>}
			(if you don't touch the recursive parameter). You can also
			specify the path deeper into the tree, i.e. I{/NetObjects/work}
			will list all L{NetObjects <netobject.NetObject>} with the object I{work} as parent.
		@type path: C{str}

		@param recursive: This specifies if the list should be recursively done.
			When set to False it will only list the items on the current
			level of path.
		@type recursive: C{boolean}

		@param systemid: If specified limits the list to only include objects
		                 with the specified systemid.
		@type systemid: C{str}

		@returns: A list of L{plobject.PLObject} subclassed objects.
		@rtype: C{list} of L{plobject.PLObject}

		@raises RuntimeError: B{"Unknown object item type"}:
			for invalid type of object. Should never happen unless
			e.g two api-versions are mixed up.
		"""

		(baseobj, pathitems) = self._split_pathstr(path)

		if baseobj == None:
			objlist = []
			for obj in self.objects.values():
				self._update_cache(obj)
				objlist += self._objectcaches[obj].list(recursive=recursive, systemid=systemid) ###
		else:
			self._update_cache(self.objects[baseobj])
			objlist = self._objectcaches[self.objects[baseobj]].list(path, recursive=recursive, systemid=systemid) ###

		return objlist

	def object_find(self, path, systemid=None):
		"""
		This method finds an object by type and name.

			>>> r.object_find('/NetObjects/TestObject')
			<PacketLogic NetObject 'TestObject' '/NetObjects'>

			>>> print r.object_find('/NetObjects/DoesNotExist')
			None

		It is a synonym for object_get, that soon will be deprecated.

		@type path: C{str}
		@param path: a path string (i.e. /NetObjects/Corporate)

		@param systemid: Get object with specified systemid if specified,
		                 otherwise default to local object.
		@type systemid: C{str}

		@returns: Object if found, else None
		@rtype: L{plobject.PLObject}

		@raises RuntimeError: B{"Unknown object item type"}:
			for invalid type of object. Should never happen unless
			e.g two api-versions are mixed up.
		"""

		(o, p) = self._split_pathstr_check(path)
		t = self.objects[o]
		self._update_cache(t)

		return self._objectcaches[t].find_path(path, systemid) ###
		
	def object_get(self, path, systemid=None):
		"""
		This method gets an object by type and name.

			>>> r.object_get('/NetObjects/TestObject')
			<PacketLogic NetObject 'TestObject' '/NetObjects'>

			>>> r.object_get('/NetObjects/TestObject')
			<PacketLogic NetObject 'TestObject' '/NetObjects'>

		It will soon be deprecated in favour for the synonym object_find.
		
		@type path: C{str}
		@param path: a path string (i.e. /NetObjects/Corporate)

		@param systemid: Get object with specified systemid if specified,
		                 otherwise default to local object.
		@type systemid: C{str}

		@returns: Object if found, else None
		@rtype: L{plobject.PLObject}

		@raises RuntimeError: B{"Unknown object item type"}:
			for invalid type of object. Should never happen unless
			e.g two api-versions are mixed up.
		"""

		(o, p) = self._split_pathstr_check(path)

		t = self.objects[o]
		self._update_cache(t)
		return self._objectcaches[t].find_path(path, systemid) ###

	def object_find_id(self, path, object_id, systemid=None):
		"""
		This method finds a object by type and id number.

			>>> r.object_find_id('/NetObjects', 31573)
			<PacketLogic NetObject 'TestObject' '/NetObjects'>

			>>> print r.object_find_id('/NetObjects', 7777777)
			None

		@type path: C{str}
		@param path: a path string (i.e. /NetObjects)

		@type object_id: C{int}
		@param object_id: ID number of the Object. Could be found by
			examining the id property of any L{plobject.PLObject} implementation.

		@param systemid: Get object with specified systemid if specified,
		                 otherwise default to local object.
		@type systemid: C{str}

		@returns: A L{plobject.PLObject} subclass.
		@rtype: L{plobject.PLObject}

		@raises RuntimeError: B{"Unknown object item type"}:
			for invalid type of object. Should never happen unless
			e.g two api-versions are mixed up.
		"""

		(o, p) = self._split_pathstr_check(path)

		t = self.objects[o]
		self._update_cache(t)
		return self._objectcaches[t].find_id(object_id, systemid) ###

	def object_add(self, path):
		"""
		Add an Object to the object-tree. The type and name of
		the Object is determined by the path.

		@param path: The path is the new Object to be created. if we want to
			create a NetObject called I{Anders} under with the NetObject
			I{Work} as parent you use the path I{/NetObjects/Work/Anders}.
		@type path: C{str}

		@returns: A L{plobject.PLObject} subclass.
		@rtype: L{plobject.PLObject}
		
		@raises RuntimeError: B{"Unknown object item type"}:
			for invalid type of object. Should never happen unless
			e.g two api-versions are mixed up.
		"""

		(obj, p) = self._split_pathstr_check(path, allowEmpty=False)

		# we add to the cache later, make sure it exists
		t = self.objects[obj]
		self._update_cache(t)

		path = str.join("/", p[:-1])
		name = p[-1]
		
		parent = 0
		parent_name = "/" + obj
		if path:
			t = self.object_get("/" + obj + "/" + path)
			if t is None:
				raise ValueError("Parent object '%s' not found" % ("/" + obj + "/" + path))
			parent = t.id
			parent_name = t.fullpath

		res = plobject._add(self, self.objects[obj], name, parent, parent_name)
		
		# the systemid will be wrong when proxying...
		self._objectcaches[res.type].add(res) ###
		
		return res

	def add(self, path):
		"""
		Add a Object to the object-tree. The type and name of
		the Object is determined by the path.

		@param path: The path is the new Object to be created. if we want to
			create a NetObject called I{Anders} under with the NetObject
			I{Work} as parent you use the path I{/NetObjects/Work/Anders}.
		@type path: C{str}

		@returns: id of the newly created object
		@rtype: C{int}

		@deprecated: Port your code to L{object_add} which differs slightly
			in return value.
		"""
		import warnings
		warnings.warn("Use Ruleset.object_add() instead of Ruleset.add(). It returns object instead of id, but is otherwise the same.",
			      DeprecationWarning, stacklevel=2)
		
		res = self.object_add(path)
		return res and res.id

	def remove(self, path):
		"""
		Removes an Object and all its children.

		@param path: The Object to be removed.
		@type path: C{str}

		@deprecated: Renamed to L{object_remove}
			Using this method issues a deprecation warning after 11.0beta5.
		"""
		import warnings
		warnings.warn("Ruleset.remove() is deprecated, use Ruleset.object_remove() instead.",
			      DeprecationWarning, stacklevel=2)
		
		self._object_remove(path)

	def object_remove(self, object):
		"""
		Removes an Object and all its children.

		Note that object_unlink only unlinks the current object, while
		object_remove also removes children. To both unlink and remove
		and object, you will need to call unlink for all children too.

			>>> r.object_add("/NetObjects/Example")
			>>> r.object_remove("/NetObjects/Example")

			>>> obj = r.object_add("/NetObjects/Example")
			>>> r.object_remove(obj)

		@param object: The Object to be removed or the full path.
		@type object: C{str} or L{plobject.PLObject}

		"""

		if isinstance(object, plobject.PLObject):
			self._object_remove(object.fullpath)
		else:
			self._object_remove(object)

	def _objectcache_move(self, obj, new_parent):
		self._objectcaches[obj.type].move(obj, new_parent)
	
	def _object_remove(self, path):
		obj = self.object_get(path)
		if obj is None:
			raise ValueError("Object '%s' not found" % path)

		plobject._remove(self, obj)
		self._objectcaches[obj.type].remove(obj) ###

	def unlink(self, path):
		"""
		Removes an Object for all firewall and shapingrules.

		This method will soon be deprecated. Use synonym object_unlink instead.

		@type path: C{str}
		@param path: The Object to be removed.

		"""
		return self.object_unlink(path)
	
	def object_unlink(self, path):
		"""
		Removes an Object for all firewall and shapingrules.
		
		Note that object_unlink only unlinks the current object, while
		object_remove also removes children. To both unlink and remove
		and object, you will need to call unlink for all children too.

		This method replaces unlink that will soon be deprecated.

		@type path: C{str}
		@param path: The Object to be removed.

		"""

		obj = self.object_get(path)
		if obj is None:
			raise ValueError("Object '%s' not found" % path)

		plobject._unlink(self, obj)
		self._fwrule_invalidate_cache()
		self._shapingrule_invalidate_cache()


	#
	# fwrule
	#

	def _fwrule_update_cache(self):
		if self._fwrulecache:
			return
		
		weighted = [(r.orderby, r.position, r) for r in plrule._fwlist(self)]
		weighted.sort()
		l = [r[-1] for r in weighted]
		
		self._fwrulecache = l
		self._fwrulecache_id = dict([((r.systemid, r.id), r) for r in l])
		self._fwrulecache_id.update(dict([((None, r.id), r) for r in l if not r.is_global]))
		self._fwrulecache_name = dict([((r.systemid, r.name), r) for r in l])
		self._fwrulecache_name.update(dict([((None, r.name), r) for r in l if not r.is_global]))

	def _fwrule_invalidate_cache(self):
		self._fwrulecache = None
		self._fwrulecache_id = None
		self._fwrulecache_name = None

	def fwrule_list(self):
		"""
		Return a list of all firewall rules.

			>>> r.fwrule_list()
			[<PacketLogic Rule '<PacketLogic Rule 'allow all'>, <PacketLogic Rule 'nxs_accept'>, <PacketLogic Rule 'Allow all DHCP'>, <PacketLogic Rule 'Netaccess redirect'>, <PacketLogic Rule 'Netaccess DHCP'>, <PacketLogic Rule 'Netaccess'>]

		@returns: A list of firewall rules.
		@rtype: C{list} of L{FwRule <plrule.FwRule>}
		"""
		self._fwrule_update_cache()
		return self._fwrulecache[:]

	def fwrule_add(self, name, action, quick=False, log=False, rewrite_object=0, monitor_iface=0, inject_data="", divert_iface=0):
		"""
		Create a new firewall rule.

			>>> r.fwrule_add('Test', r.FWRULE_ACTION_ACCEPT, quick=True)
			<PacketLogic Rule 'Test'>
		
			>>> rewrite_object_id = r.object_get('/RewriteObjects/Netaccess server').id
			>>> r.fwrule_add('Test', r.FWRULE_ACTION_REWRITE, quick=True, rewrite_object=rewrite_object_id)
			<PacketLogic Rule 'Test'>
		
			>>> rewrite_object = r.object_get('/RewriteObjects/Netaccess server')
			>>> r.fwrule_add('Test', r.FWRULE_ACTION_REWRITE, quick=True, rewrite_object=rewrite_object)
			<PacketLogic Rule 'Test'>

		@type name: C{str}
		@param name: The name of the new firewall rule, a rule
		with this name must not already exist.

		@type action: C{int}
		@param action: Action when the rule matches, one of:
		FWRULE_ACTION_ACCEPT FWRULE_ACTION_REJECT,
		FWRULE_ACTION_DROP, FWRULE_ACTION_REWRITE,
		FWRULE_ACTION_DIVERT, FWRULE_ACTION_INJECT.

		@type quick: C{boolean}
		@param quick: Give this rule the quick property.

		@type log: C{int}
		@param log: Log level to use for this rule. 0 = Off. 1 = Brief. 2 = Verbose

		@type rewrite_object: C{int} or L{rewriteobject.RewriteObject}
		@param rewrite_object: The id or the actual L{rewriteobject.RewriteObject}
		describing what to rewrite when this rule matches and action is
		FWRULE_ACTION_REWRITE.

		@type monitor_iface: C{int}
		@param monitor_iface: If non zero, packets matching
		this rule will also be sent to the specified
		interface id. See L{monitor_iface_list <Ruleset.monitor_iface_list>}.

		@type inject_data: C{str}
		@param inject_data: The content that shall be injected for
		FWRULE_ACTION_INJECT rules. The content is usually a HTTP 307 
		redirect response.

		@type divert_iface: C{int}
		@param divert_iface: Interface to divert packets to when
		using FWRULE_ACTION_DIVERT. See L{divert_iface_list <Ruleset.divert_iface_list>}.

		@returns: Created firewall rule.
		@rtype: L{FwRule <plrule.FwRule>}

		@raises PLDBUnsupportedInFirmware: If inject_data/divert_iface arguments are used and firmware doesn't support them.

		@attention: inject_data parameter requires v12.1 firmware or newer.
		@attention: divert_iface parameter requires v12.2 firmware or newer.
		"""

		if inject_data:
			raise PLDBUnsupportedInFirmware("Usage of 'inject_data' is not supported in this firmware version")

		if divert_iface:
			raise PLDBUnsupportedInFirmware("Usage of 'divert_iface' is not supported in this firmware version")

		if action is plrule.NONEXISTANT_FWRULE_ACTION:
			raise PLDBUnsupportedInFirmware("Specified action is not supported in this firmware version")

		if log not in [0,1,2]:
			raise ValueError("Unknown log level")

		# we add to the cache later, make sure it exists
		self._fwrule_update_cache()

		if isinstance(rewrite_object, rewriteobject.RewriteObject):
			rewrite_object_id = rewrite_object.id
		else:
			rewrite_object_id = rewrite_object

		res = plrule._fwadd(self, name, action, quick, int(log), rewrite_object_id, monitor_iface)

		if len(self._fwrulecache) == 0 or (res.position > self._fwrulecache[-1].position and res.orderby == self._fwrulecache[-1].orderby):
			self._fwrulecache.append(res)
		else:
			# This usually never happens, but in a multi-system environment it just might...
			for ix, r in enumerate(self._fwrulecache):
				if r.orderby > res.orderby or (r.orderby == res.orderby and r.position > res.position):
					# We want to be just before this one, i.e take it's place
					self._fwrulecache[ix:ix] = [res]
					break
			else:
				# This should have been handled by the outer if... but you never know in the computer world :)
				self._fwrulecache.append(res)
			
		# this systemid will be wrong when proxying...
		self._fwrulecache_id[(res.systemid, res.id)] = res
		self._fwrulecache_id[(None, res.id)] = res
		# this systemid will be wrong when proxying...
		self._fwrulecache_name[(res.systemid, res.name)] = res
		self._fwrulecache_name[(None, res.name)] = res

		return res

	def fwrule_find(self, name, systemid=None):
		"""
		Find a firewallrule by name

			>>> r.fwrule_find("allow all")
			<PacketLogic Rule 'allow all'>

		@type name: C{str}
		@param name: The name of the firewall rule.

		@returns: Firewall rule or None if not found.
		@rtype: L{FwRule <plrule.FwRule>}
		"""
		self._fwrule_update_cache()
		return self._fwrulecache_name.get((systemid, name), None)

	def fwrule_find_id(self, rule_id, systemid=None):
		"""
		Find a firewallrule by id

			>>> r.fwrule_find_id(14)
			<PacketLogic Rule 'allow all'>

		@type rule_id: C{int}
		@param rule_id: The id of the firewall rule.

		@returns: Firewall rule or None if not found.
		@rtype: L{FwRule <plrule.FwRule>}
		"""
		self._fwrule_update_cache()
		return self._fwrulecache_id.get((systemid, rule_id), None)

	def fwrule_remove(self, rule):
		"""
		Remove firewall rule

			>>> r.fwrule_add('Example', r.FWRULE_ACTION_ACCEPT, quick=True)
			<PacketLogic Rule 'Example'>
			>>> r.fwrule_remove('Example')

			>>> rule = r.fwrule_add('Example', r.FWRULE_ACTION_ACCEPT, quick=True)
			>>> r.fwrule_remove(rule)

		@param rule: The name or actual rule to be removed.
		@type rule: C{str} or L{plrule.FwRule}

		@raises ValueError: If firewall rule is not found
		"""
		if isinstance(rule, plrule.FwRule):
			name = rule.name
		else:
			name = rule

		obj = self.fwrule_find(name)

		if obj is None:
			raise ValueError("Fwrule '%s' does not exist" % name)

		plrule._fwremove(self, obj)

		self._fwrulecache.remove(obj)
		del self._fwrulecache_id[(obj.systemid, obj.id)]
		del self._fwrulecache_id[(None, obj.id)]
		del self._fwrulecache_name[(obj.systemid, obj.name)]
		del self._fwrulecache_name[(None, obj.name)]

	#
	# monitor interfaces
	#

	def monitor_iface_list(self):
		"""
		Return a list of all monitor interfaces.

			>>> r.monitor_iface_list()
			[<PacketLogic Monitor Interface 'PCAP Writer'>, <PacketLogic Monitor Interface 'DHCP Snooper'>, <PacketLogic Monitor Interface 'RADIUS Snooper'>, <PacketLogic Monitor Interface 'PCAP-2 Writer'>]

		@returns: A list of monitor interfaces.
		@rtype: read-only C{list} of L{MonitorIface <plrule.MonitorIface>}
		"""
		return plutil.RoList(plrule._monitoriface_list(self))

	#
	# divert interfaces
	#

	def divert_iface_list(self):
		"""
		Return a list of all divert interfaces.

			>>> r.monitor_iface_list()
			[]

		@returns: A list of divert interfaces.
		@rtype: read-only C{list} of L{DivertIface <plrule.DivertIface>}
			
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.2 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("divert_iface_list requires firmware v12.2 or newer")

	#
	# shapingrule
	#

	def _shapingrule_update_cache(self):
		if self._shapingrulecache:
			return
		l = plrule._shapinglist(self)
		self._shapingrulecache = l
		self._shapingrulecache_id = dict([((r.systemid, r.id), r) for r in l])
		self._shapingrulecache_id.update(dict([((None, r.id), r) for r in l if not r.is_global]))
		self._shapingrulecache_name = dict([((r.systemid, r.name), r) for r in l])
		self._shapingrulecache_name.update(dict([((None, r.name), r) for r in l if not r.is_global]))

	def _shapingrule_invalidate_cache(self):
		self._shapingrulecache = None
		self._shapingrulecache_id = None
		self._shapingrulecache_name = None

	def shapingrule_list(self):
		"""
		Return a list of all shaping rules.

			>>> r.shapingrule_list()
			[<PacketLogic Rule 'Example'>, <PacketLogic Rule 'Example2'>]

		@returns: A list of L{ShapingRules <plrule.ShapingRule>}
		@rtype: C{list} of L{ShapingRule <plrule.ShapingRule>}
		"""
		self._shapingrule_update_cache()
		return self._shapingrulecache[:]

	def shapingrule_find(self, name, systemid=None):
		"""
		Find a shapingrule by name

			>>> r.shapingrule_find('Example')
			<PacketLogic Rule 'Example'>

		@type name: C{str}
		@param name: Name of shaping rule to find

		@returns: L{ShapingRule <plrule.ShapingRule>}, or None if not found.
		@rtype: L{ShapingRule <plrule.ShapingRule>}
		"""
		self._shapingrule_update_cache()
		return self._shapingrulecache_name.get((systemid, name), None)

	def shapingrule_find_id(self, rule_id, systemid=None):
		"""
		Find a shapingrule by id

			>>> r.shapingrule_find(27)
			<PacketLogic Rule 'Example'>

		@type rule_id: C{int}
		@param rule_id: Id of shaping rule to find
		
		@returns: L{ShapingRule <plrule.ShapingRule>} or None if not found
		@rtype: L{ShapingRule <plrule.ShapingRule>}
		"""
		self._shapingrule_update_cache()
		return self._shapingrulecache_id.get((systemid, rule_id), None)

	def shapingrule_add(self, name, priority=5, max_connections=None):
		"""
		Create a new shapingrule in pldb

			>>> r.shapingrule_add('Example')
			<PacketLogic Rule 'Example'>

			>>> r.shapingrule_add('Example2', priority=20)
			<PacketLogic Rule 'Example2'>

		@type name: C{str}
		@param name: Name of the new rule

		@type priority: C{int}
		@param priority: Priority of packets matching this rule

		@return: Newly created shaping rule
		@rtype: L{ShapingRule <plrule.ShapingRule>}

		@deprecated: The max_connections argument is not supported since version 11.x
			Supplying it as 0 will generate a deprecation warning, a nonzero value will raise an exception.
			Use max_connections argument on L{shapingobject_add() <Ruleset.shapingobject_add>} instead.
		"""

		# Check if we supply the obsolete/deprecated argument max_connections
		if max_connections == 0:
			import warnings
			warnings.warn("Use of argument 'max_connections' on shapingrule_add() is deprecated.",
				      DeprecationWarning, stacklevel=2)

		elif max_connections is not None:
			raise TypeError("shapingrule_add() got an unexpected keyword argument 'max_connections'")
			

		if not isinstance(priority, int):
			raise TypeError

		if priority < 0 or priority >255:
			raise ValueError

		# we add to the cache later, make sure it exists
		self._shapingrule_update_cache()

		res = plrule._shapingadd(self, name, priority)

		self._shapingrulecache.append(res)
		# this systemid will be wrong when proxying...
		self._shapingrulecache_id[(res.systemid, res.id)] = res
		self._shapingrulecache_id[(None, res.id)] = res
		# this systemid will be wrong when proxying...
		self._shapingrulecache_name[(res.systemid, res.name)] = res
		self._shapingrulecache_name[(None, res.name)] = res

		return res

	def shapingrule_remove(self, rule):
		"""
		Remove shaping rule

			>>> r.shapingrule_add('Example')
			<PacketLogic Rule 'Example'>
			>>> r.shapingrule_remove('Example')

			>>> rule = r.shapingrule_add('Example')
			>>> r.shapingrule_remove(rule)

		@param rule: The name or actual rule to be removed.
		@type rule: C{str} or L{plrule.ShapingRule}

		@raise ValueError: If shaping rule is not found

		"""
		if isinstance(rule, plrule.ShapingRule):
			name = rule.name
		else:
			name = rule

		obj = self.shapingrule_find(name)

		if obj is None:
			raise ValueError("Shapingrule '%s' does not exist" % name)

		plrule._shapingremove(self, obj)

		self._shapingrulecache.remove(obj)
		del self._shapingrulecache_id[(obj.systemid, obj.id)]
		del self._shapingrulecache_id[(None, obj.id)]
		del self._shapingrulecache_name[(obj.systemid, obj.name)]
		del self._shapingrulecache_name[(None, obj.name)]

	#
	# statisticsrule
	#

	def statisticsrule_list(self):
		"""
		Return a list of all statistics rules.

			>>> r.statisticsrule_list()
			[<PacketLogic Rule 'Example'>, <PacketLogic Rule 'Example2'>]

		@returns: A list of L{StatisticsRules <plrule.StatisticsRule>}
		@rtype: C{list} of L{StatisticsRule <plrule.StatisticsRule>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsrule requires firmware v12.0 or newer")

	def statisticsrule_find(self, name, systemid=None):
		"""
		Find a statisticsrule by name

			>>> r.statisticsrule_find('Example')
			<PacketLogic Rule 'Example'>

		@type name: C{str}
		@param name: Name of statistics rule to find

		@returns: L{StatisticsRule <plrule.StatisticsRule>} or None if not found
		@rtype: L{StatisticsRule <plrule.StatisticsRule>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsrule requires firmware v12.0 or newer")

	def statisticsrule_find_id(self, rule_id, systemid=None):
		"""
		Find a statisticsrule by id

			>>> r.statisticsrule_find(27)
			<PacketLogic Rule 'Example'>

		@type rule_id: C{int}
		@param rule_id: Id of statistics rule to find
		
		@returns: L{StatisticsRule <plrule.StatisticsRule>} or None if not found
		@rtype: L{StatisticsRule <plrule.StatisticsRule>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsrule requires firmware v12.0 or newer")

	def statisticsrule_add(self, name):
		"""
		Create a new statisticsrule in pldb

			>>> r.statisticsrule_add('Example')
			<PacketLogic Rule 'Example'>

			>>> r.statisticsrule_add('Example2', priority=20)
			<PacketLogic Rule 'Example2'>

		@type name: C{str}
		@param name: Name of the new rule

		@return: Newly created statistics rule
		@rtype: L{StatisticsRule <plrule.StatisticsRule>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsrule requires firmware v12.0 or newer")

	def statisticsrule_remove(self, rule):
		"""
		Remove statistics rule

			>>> r.statisticsrule_add('Example')
			<PacketLogic Rule 'Example'>
			>>> r.statisticsrule_remove('Example')

			>>> rule = r.statisticsrule_add('Example')
			>>> r.statisticsrule_remove(rule)

		@param rule: The name or actual rule to be removed.
		@type rule: C{str} or L{plrule.StatisticsRule}

		@raise ValueError: If statistics rule is not found

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsrule requires firmware v12.0 or newer")


	#
	# shapingobject
	#

	def _shapingobject_update_cache(self):
		if self._shapingobjectcache:
			return
		l = shapingobject._shapingobjectlist(self)
		self._shapingobjectcache = l
		self._shapingobjectcache_id = dict([((r.systemid, r.id), r) for r in l])
		self._shapingobjectcache_id.update(dict([((None, r.id), r) for r in l if not r.is_global]))
		self._shapingobjectcache_name = dict([((r.systemid, r.name), r) for r in l])
		self._shapingobjectcache_name.update(dict([((None, r.name), r) for r in l if not r.is_global]))

	def _shapingobject_invalidate_cache(self):
		self._shapingobjectcache = None
		self._shapingobjectcache_id = None
		self._shapingobjectcache_name = None

	def shapingobject_list(self):
		"""
		Return a list of all shaping objects.

			>>> r.shapingobject_list()
			[<PacketLogic ShapingObject ('Test VBS', inbound=[(128, 0, 1024), (256, 0, 512), (512, 0, 128), (0, 0, 64)], outbound=[(0, 0, 0)], bidir=[(0, 0, 0)], split=0, vbs_hours=720)>,
			 <PacketLogic ShapingObject ('Test Shaping', inbound=[(0, 0, 1024)], outbound=[(0, 0, 512)], bidir=[(0, 0, 0)], split=0, vbs_hours=0)>]

		@returns: A list of L{ShapingObjects <shapingobject.ShapingObject>}
		@rtype: C{list} of L{ShapingObject <shapingobject.ShapingObject>}
		"""
		self._shapingobject_update_cache()
		return self._shapingobjectcache[:]

	def shapingobject_find(self, name, systemid=None):
		"""
		Find a shaping object by its name

			>>> r.shapingobject_find('Test Shaping')
			<PacketLogic ShapingObject ('Test Shaping', inbound=[(0, 0, 1024)], outbound=[(0, 0, 512)], bidir=[(0, 0, 0)], split=0, vbs_hours=0)>

		@type name: C{str}
		@param name: Name of shaping object to find

		@returns: ShapingObject or None if not found
		@rtype: L{ShapingObject <shapingobject.ShapingObject>}
		"""
		self._shapingobject_update_cache()
		return self._shapingobjectcache_name.get((systemid, name), None)

	def shapingobject_find_id(self, object_id, systemid=None):
		"""
		Find a shaping object by its id

			>>> r.shapingobject_find_id(83)
			<PacketLogic ShapingObject ('Test Shaping', inbound=[(0, 0, 1024)], outbound=[(0, 0, 0)], bidir=[(0, 0, 0)], split=1, vbs_hours=0)>

		@type object_id: C{int}
		@param object_id: Id of shaping object to find

		@returns: ShapingObject or None if not found.
		@rtype: L{ShapingObject <shapingobject.ShapingObject>}
		"""
		self._shapingobject_update_cache()
		return self._shapingobjectcache_id.get((systemid, object_id), None)

	def shapingobject_add(self, name, inbound=[(0, 0, 0)], outbound=[(0, 0, 0)], bidir=[(0, 0, 0)], split=SPLIT_NONE, vbs_hours=0, max_connections=0, flags=[]):
		"""
		Create a new L{ShapingObject <shapingobject.ShapingObject>} in pldb

			>>> r.shapingobject_add('Test Shaping',
			...	inbound=[(0, 0, 1024)], # 1024 kbyte/s inbound
			...	outbound=[(0, 0, 512)], # 512 kbyte/s outbound
			...	bidir=[(0, 0, 0)])
			<PacketLogic ShapingObject ('Test Shaping', inbound=[(0, 0, 1024)], outbound=[(0, 0, 512)], bidir=[(0, 0, 0)], split=0, vbs_hours=0)>

			>>> r.shapingobject_add('Test VBS', inbound=[
			...		(128, 0, 1024),  # 128 MiB at 1024 kbyte/s
			...		(256, 0,  512),  # 256 MiB at 512 kbyte/s
			...		(512, 0,  128),  # 512 MiB at 128 kbyte/s
			...		(  0, 0,   64)], # Unlimited at 64 kbyte/s
			...	outbound=[(0, 0, 0)],
			...	bidir=[(0, 0, 0)],
			...	vbs_hours=720) # 30 days (30*24 hours)
			<PacketLogic ShapingObject ('Test VBS', inbound=[(128, 0, 1024), (256, 0, 512), (512, 0, 128), (0, 0, 64)], outbound=[(0, 0, 0)], bidir=[(0, 0, 0)], split=0, vbs_hours=720)>

		@type name: C{str}
		@param name: Name of the new object

		@type inbound: C{list} of C{tuple}
		@param inbound: list of limits for incoming traffic,
	                        where each limit is (max_transfer,
	                        max_packets, max_speed) max_transfer
	                        must be 0 (unlimited) on the last
	                        entry in the list.

		@type outbound: C{list} of C{tuple}
	        @param outbound: same as inbound but for outgoing traffic
		
		@type bidir: C{list} of C{tuple}
	        @param bidir: same as inbound but for both directions

		@type split: C{int}
		@param split: Type of split, one of: SPLIT_NONE,
		              SPLIT_LOCALHOST, SPLIT_HOST_NETOBJECT,
		              SPLIT_LOCAL_NETOBJECT, SPLIT_SERVER_NETOBJECT,
		              SPLIT_CLIENT_NETOBJECT, SPLIT_CONNECTION,
			      SPLIT_SUBSCRIBER

		@type vbs_hours: C{int}
		@param vbs_hours: Duration in hours if this is a VBS (Volume Based Shaping) object.
				  Value 0 means that VBS is not used on this object.
				  Value 1 means that no sliding window is used, and you will need
				  to reset volume manually with L{Realtime.vbs_reset <pld.PLD.vbs_reset>}.

		@type max_connections: C{int}
		@param max_connections: Limit number of connections, or 0 for unlimited.

		@type flags: C{list} of C{str}
		@param flags: List of the flags to set, if any.
		              Flags can be any combination of: 'brown', 'host_fairness'

		@returns: ShapingObject
		@rtype: L{ShapingObject <shapingobject.ShapingObject>}
		"""

		# we add to the cache later, make sure it exists
		self._shapingobject_update_cache()

		if split is shapingobject.NONEXISTANT_SPLIT_TYPE:
			raise PLDBUnsupportedInFirmware("Specified split type is not supported in this firmware version")

		if inbound[-1][0] != 0 or outbound[-1][0] != 0 or bidir[-1][0] != 0:
			raise ValueError("last limit must have max_transfer=0")
		if vbs_hours == 0 and (len(inbound) != 1 or len(outbound) != 1 or len(bidir) != 1):
			raise ValueError("Limit list must be of length 1 when vbs_hours is 0")
		if split not in [
			self.SPLIT_NONE, self.SPLIT_LOCALHOST, self.SPLIT_HOST_NETOBJECT,
			self.SPLIT_SERVER_NETOBJECT, self.SPLIT_CLIENT_NETOBJECT,
			self.SPLIT_CONNECTION, self.SPLIT_LOCAL_NETOBJECT, self.SPLIT_SUBSCRIBER]:
			raise ValueError("Invalid split type")

		#we could probably add more sanity checking here...

		res = shapingobject._shapingobjectadd(self, name, vbs_hours, inbound, outbound, bidir, split, max_connections, flags)

		self._shapingobjectcache.append(res)
		# this systemid will be wrong when proxying...
		self._shapingobjectcache_id[(res.systemid, res.id)] = res
		self._shapingobjectcache_id[(None, res.id)] = res
		# this systemid will be wrong when proxying...
		self._shapingobjectcache_name[(res.systemid, res.name)] = res
		self._shapingobjectcache_name[(None, res.name)] = res

		return res

	def shapingobject_remove(self, object):
		"""
		Remove shaping object

			>>> r.shapingobject_add("Example")
			<PacketLogic ShapingObject ('Example', inbound=[(0, 0, 0)], outbound=[(0, 0, 0)], bidir=[(0, 0, 0)], split=0, vbs_hours=0)>
			>>> r.shapingobject_remove("Example")

			>>> object = r.shapingobject_add("Example")
			>>> r.shapingobject_remove(object)

		@param object: The name or actual rule to be removed.
		@type object: C{str} or L{plrule.FwRule}

		@raise ValueError: If shaping object is not found
		"""
		if isinstance(object, shapingobject.ShapingObject):
			name = object.name
		else:
			name = object

		obj = self.shapingobject_find(name)

		if obj is None:
			raise ValueError("Shapingobject '%s' does not exist" % name)
			
		shapingobject._remove(self, obj)

		self._shapingobjectcache.remove(obj)
		del self._shapingobjectcache_id[(obj.systemid, obj.id)]
		del self._shapingobjectcache_id[(None, obj.id)]
		del self._shapingobjectcache_name[(obj.systemid, obj.name)]
		del self._shapingobjectcache_name[(None, obj.name)]

	#
	# statisticsobject
	#

	def statisticsobject_list(self):
		"""
		Return a list of all statistics objects.

			>>> r.statisticsobject_list()
			[<PacketLogic StatisticsObject>,
			 <PacketLogic StatisticsObject>]

		@returns: A list of L{StatisticsObjects <statisticsobject.StatisticsObject>}
		@rtype: C{list} of L{StatisticsObject <statisticsobject.StatisticsObject>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsobject requires firmware v12.0 or newer")

	def statisticsobject_find(self, name, systemid=None):
		"""
		Find a statistics object by its name

			>>> r.statisticsobject_find('Test Statistics')
			<PacketLogic StatisticsObject>

		@type name: C{str}
		@param name: Name of statistics object to find

		@returns: StatisticsObject or None if not found
		@rtype: L{StatisticsObject <statisticsobject.StatisticsObject>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsobject requires firmware v12.0 or newer")

	def statisticsobject_find_id(self, object_id, systemid=None):
		"""
		Find a statistics object by its id

			>>> r.statisticsobject_find_id(83)
			<PacketLogic StatisticsObject>

		@type object_id: C{int}
		@param object_id: Id of statistics object to find

		@returns: StatisticsObject or None if not found.
		@rtype: L{StatisticsObject <statisticsobject.StatisticsObject>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsobject requires firmware v12.0 or newer")


	def statisticsobject_add(self, name, total_fields, graph_fields, split, no_root=0, so_root=0, graph_freq=300, min_bytes_in=102400, min_bytes_out=102400):
		"""
		Create a new L{StatisticsObject <statisticsobject.StatisticsObject>} in pldb

			>>> s = pl.Statistics()
			>>> r.statisticsobject_add('Test Statistics', 
			    s.TOTALFIELD_BYTES_IN | s.TOTALFIELD_BYTES_OUT,
			    s.GRAPHFIELD_BPS_IN | s.GRAPHFIELD_BPS_OUT | s.GRAPHFIELD_CPS,
			    "(NEO:2(SEO))")
			<PacketLogic StatisticsObject 'Test Statistics'>

		@type name: C{str}
		@param name: Name of the new object

		@type total_fields: C{int}
		@param total_fields: Bitmask of fields to store total values for.

		@type graph_fields: C{int}
		@param graph_fields: Bitmask of fields to store graphs for.

		@type split: C{str}
		@param split: String describing distribution to use.

		@type no_root: C{int}
		@param no_root: ID of NetObject to use as root.

		@type so_root: C{int}
		@param so_root: ID of ServiceObject to use as root.

		@type graph_freq: C{int}
		@param graph_freq:

		@type min_bytes_in: C{int}
		@param min_bytes_in: Threshold for inbound values. Values below this threshold are not stored.

		@type min_bytes_out: C{int}
		@param min_bytes_out: Threshold for outbound values. Values below this threshold are not stored.

		@returns: StatisticsObject
		@rtype: L{StatisticsObject <statisticsobject.StatisticsObject>}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsobject requires firmware v12.0 or newer")

	def statisticsobject_remove(self, object):
		"""
		Remove statistics object

			>>> r.statisticsobject_add("Example")
			<PacketLogic StatisticsObject ('Example')>
			>>> r.statisticsobject_remove("Example")

			>>> object = r.statisticsobject_add("Example")
			>>> r.statisticsobject_remove(object)

		@param object: The name or actual rule to be removed.
		@type object: C{str} or L{statisticsobject.StatisticObject}

		@raise ValueError: If statistics object is not found
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		raise PLDBUnsupportedInFirmware("statisticsobject requires firmware v12.0 or newer")
