"""
This module contains DSCPObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
DSCPObjects.

    >>> o = r.object_add("/DSCPObjects/Test")
    >>> o.add("5")
    >>> o.add("1-3")
    >>> print o.items
    ['5', '1-3']

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plobject
from thisversion import PLDB
import plutil

def _verify_dscp(dscp_id):
	if dscp_id < 0 or dscp_id > (1 << 6):
		return False
	return True

OBJECT_TYPE = -1

class DSCPObject(plobject.PLObject):
	"""
	dscpObject Implementation

	Object path: /dscpObjects

	@cvar Item: dscpObject Item implementation.
	"""

	__slots__ = ('_prop', '_pldb', '_callback', '_item_added')
	_may_exclude = True

	def __init__(self, **props):
		plobject.PLObject.__init__(self, **props)

	base_path = plutil.roConst("/DSCPObjects", "Only the type portion of the path")
	type = plutil.roConst(OBJECT_TYPE, "Number representing the type of object")

	def _create_item(self, value):
		if isinstance(value, basestring) and "-" in value:
			# type range!
			first, last = value.split("-", 1)
			try:
				item = self.Item(self, type=PLDB.DSCPOBJECT_ITEM_TYPE_RANGE, value1=first, value2=last)
			except:
				raise ValueError("Format is not correct for this item type")

			if not _verify_dscp(item._prop.value1) or not _verify_dscp(item._prop.value2):
				raise ValueError("dscp id must be between 0 and 64")
			
			if item._prop.value1 > item._prop.value2:
				raise ValueError("Value2 is smaller than Value1")

		else:
			# type address
			item = self.Item(self, type=PLDB.DSCPOBJECT_ITEM_TYPE_DSCP, value1=value, value2=0)

			if not _verify_dscp(item._prop.value1):
				raise ValueError("First argument doesn't seem like a dscp id")

		return item

	def add(self, value, exclude=False):
		"""
		Adds a dscpObject item to this dscpObject.

		@type value: C{str}
		@param value: The value is a string describing the dscpObject item. The type of the item
			will be automatically detected from the string. The following syntax is valid:
				- dscp ID, just give the plain dscp id I{XXXX}
				- dscp ID range, give the start and stop dscp id with a dash as delimiter I{XXXX-YYYY}

		@type exclude: C{boolean}
		@param exclude: If this is set to True the item will be a exclude
			Item instead of a include Item.

		@rtype: C{None}
		"""
		plobject.PLObject.add(self, value, exclude)


	class Item(plobject.PLObject.Item):
		"""
		dscpObject Item implementation.
		Should not be instantiated, use L{dscpObject.add} instead.

		@type exclude: C{boolean}
		@type value1: C{int}
		@type value2: C{int}
		"""
		__slots__ = ('_prop', '_parent')
		_Prop = plutil.Prop(exclude=bool, type=int, value1=int, value2=int, *plobject._item_props)

		def __str__(self):
			if self._prop.exclude:
				ex = "!"
			else:
				ex = ""

			if self._prop.type == PLDB.DSCPOBJECT_ITEM_TYPE_DSCP:
				return ex + str(self._prop.value1)
			elif self._prop.type == PLDB.DSCPOBJECT_ITEM_TYPE_RANGE:
				return ex + "%s-%s" % (self._prop.value1, self._prop.value2)
			else:
				raise ValueError("Wrong item-type for this object")

		type = property(lambda self: self._prop.type, doc="Internal number representing the item type (single or range)")
		exclude = property(lambda self: self._prop.exclude, doc="True = exclude, False = include")
		value1 = property(lambda self: self._prop.value1, doc="dscp ID, or first ID in a range")
		value2 = property(lambda self: self._prop.value2, doc="0 or last part of the dscp range")
