"""
This module contains functionality for managing users on the
PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> u = pl.Users()

"""

from resource import Resource
from thisversion import PLDB

import plutil


_Prop_User = plutil.Prop('id', 'name', 'readbm', 'writebm', 'survbm', 'flags', hosts=list)
class User:
    """
    Represents a user on the PacketLogic system.

    Returned from L{Users.list} and L{Users.add}.
    """
    __slots__ = ('_prop', '_pldb')
    
    def __init__(self, _pldb=None, **props):
        """
        Should not be instantiated manually, use L{list <Users.list>}.
        """
        self._pldb = _pldb
        self._prop = _Prop_User(**props)
        
    def __repr__(self):
        return "<User '%s' (#%d)>" % (self.name, self.id)

    def update(self, name=None, password=None, readbm=None, writebm=None, survbm=None, flags=None, hosts=None):
        """
        Update an attribute of this user.

        Omitting an argument or setting it to C{None} will leave that attribute unchanged.

        @type name: C{str}
        @param name: Modify user's name (rename user).

        @type password: C{str}
        @param password: Modify user's password.

        @type readbm: C{int}
        @param readbm: Modify user's readbm.

        @type writebm: C{int}
        @param writebm: Modify user's writebm.

        @type survbm: C{int}
        @param survbm: Modify user's survbm.

        @type flags: C{int}
        @param flags: Modify user's flags

        @type hosts: C{list} of C{str}
        @param hosts: Modify user's host access list
        """
        if name is None:
            name = self._prop.name
        if readbm is None:
            readbm = self._prop.readbm
        if writebm is None:
            writebm = self._prop.writebm
        if survbm is None:
            survbm = self._prop.survbm
        if flags is None:
            flags = self._prop.flags
        if hosts is None:
            hosts = list(self._prop.hosts)
        if password is None:
            password = "" # this is special as it isn't listed.

        plutil._do(self._pldb, self._pldb.session.user_modify(self._prop.id, name, password, readbm, writebm, survbm, flags, hosts), plutil._AddCB(self._pldb))

        self._prop.name = name
        self._prop.readbm = readbm
        self._prop.writebm = writebm
        self._prop.survbm = survbm
        self._prop.flags = flags
        self._prop.hosts = hosts

    def remove(self):
        """
        Remove this user from the system.
        """
        plutil._do(self._pldb, self._pldb.session.user_remove(self._prop.id), plutil._AddCB(self._pldb))

    id = property(lambda self: self._prop.id, doc="ID Number of the user.")
    name = property(lambda self: self._prop.name, doc="Name of the user.")
    readbm = property(lambda self: self._prop.readbm, doc="Bitmask of read permissions.")
    writebm = property(lambda self: self._prop.writebm, doc="Bitmask of write permissions.")
    survbm = property(lambda self: self._prop.survbm, doc="Bitmask of liveview permissions.")
    flags = property(lambda self: self._prop.flags, doc="Bitmask of flags.")
    hosts = property(lambda self: plutil.RoList(self._prop.hosts), doc="hosts user is allowed to connect from.")


class Users(Resource):
    """
    This is the resource for manipulating users.

    @group Users: list, add
    """
	
    def _resource(self):
        return PLDB.RESOURCE_USERS
    
    def list(self):
        """
        Lists all users on the system

            >>> u.list()
            [<User 'admin' (#1)>, <User 'packetlogicd' (#2)>]

        @rtype: C{list} of L{User}
        @returns: A list of all users on the system, each user is represented by an L{User} object.
        """
        
        r = plutil._do(self, self.session.user_list(),
                       plutil._ListCB(self, self.session.user_list_handle, None, False))
        return [User(self, **x) for x in r]

    def add(self, name, password, readbm=0, writebm=0, survbm=0, flags=0, hosts=[]):
        """
        Adds a new user to the system.

        @type name: C{str}
        @param name: The new user's name (rename user).

        @type password: C{str}
        @param password: The new user's password.

        @type readbm: C{int}
        @param readbm: The new user's readbm.

        @type writebm: C{int}
        @param writebm: The new user's writebm.

        @type survbm: C{int}
        @param survbm: The new user's survbm.

        @type flags: C{int}
        @param flags: The new user's flags

        @type hosts: C{list} of C{str}
        @param hosts: The new user's host access list.

        @rtype: L{User}
        @returns: An L{User} object representing the newly created user.
        """

        _id = plutil._do(self, self.session.user_add(name, password, readbm, writebm, survbm, flags, hosts), plutil._AddCB(self, self.session.user_add_handle))

        return User(self, id=_id, name=name, readbm=readbm, writebm=writebm, survbm=survbm, flags=flags, hosts=hosts)

    def auth(self, username, password, ipaddress):
        """
        Try to authenticate a user.

        The PacketLogic system will check if the the specified
        credentials matches a valid user. Note that this does NOT
        change current user or anything like that.

        @rtype: C{tuple}
        @returns: A tuple of (status, readbm, writebm, survbm) of that user.

        @raises PLDBError: If authentication fails.
        """
        return plutil._do(self, self.session.user_auth(username, password, ipaddress), plutil._AddCB(self, self.session.user_auth_handle))
